class SchemaAlreadyExistsException(Exception):
    pass


class JobFailedException(Exception):
    def __init__(self, message, data):
        self.message = message
        self.data = data


class UnableToFetchJobStatusException(Exception):
    def __init__(self, message, data):
        self.message = message
        self.data = data


class DataFrameUploadFailedException(Exception):
    def __init__(self, message, data):
        self.message = message
        self.data = data


class DataFrameUploadValidationException(Exception):
    pass


class DatasetInfoFailedException(Exception):
    def __init__(self, message, data):
        self.message = message
        self.data = data


class DatasetNotFoundException(Exception):
    def __init__(self, message, data):
        self.message = message
        self.data = data


class SchemaGenerationFailedException(Exception):
    def __init__(self, message, data):
        self.message = message
        self.data = data


class SchemaCreateFailedException(Exception):
    def __init__(self, message, data):
        self.message = message
        self.data = data


class SchemaUpdateFailedException(Exception):
    def __init__(self, message, data):
        self.message = message
        self.data = data


class CannotFindCredentialException(Exception):
    pass


class AuthenticationErrorException(Exception):
    pass


class SchemaInitialisationException(Exception):
    pass


class ColumnNotDifferentException(Exception):
    pass


class InvalidPermissionsException(Exception):
    pass


class SubjectAlreadyExistsException(Exception):
    pass


class SubjectNotFoundException(Exception):
    pass


class SubjectDeletionFailedException(Exception):
    pass



from enum import Enum
from typing import Dict, List, Optional, Union
from pydantic.main import BaseModel


class SensitivityLevel(Enum):
    PUBLIC = "PUBLIC"
    PRIVATE = "PRIVATE"
    PROTECTED = "PROTECTED"


class UpdateBehaviour(Enum):
    APPEND = "APPEND"
    OVERWRITE = "OVERWRITE"


class Owner(BaseModel):
    name: str
    email: str


class SchemaMetadata(BaseModel):
    layer: str
    domain: str
    dataset: str
    sensitivity: SensitivityLevel
    owners: List[Owner]
    version: Optional[int] = None
    key_value_tags: Optional[Dict[str, str]] = {}
    key_only_tags: Optional[List[str]] = []

    class Config:
        use_enum_values = True


class Column(BaseModel):
    name: str
    data_type: str
    partition_index: Optional[int] = None
    allow_null: bool = True
    format: Optional[str] = None


class Schema(BaseModel):
    """
    A Schema is a Pydantic class representing a rAPId schema. It allows you to programmatically define
    a schema to generate, create and update within rAPId.

    Example:
        A Schema can be created by setting the values literally into the classes like
        example below::

            schema = Schema(
                metadata=SchemaMetadata(
                    layer='default',
                    domain="domain",
                    dataset="dataset",
                    sensitivity=SensitivityLevel.PUBLIC,
                    owners=[Owner(name="test", version="test@email.com")]
                ),
                columns=[
                    Column(
                        name="column_a",
                        data_type="Float64",
                        allow_null=True
                    )
                ]
            )

        The alternative is you can create a schema directly from a Python dictionary
        specifying the values like in the example below::

            schema = Schema(
                **{
                    "metadata": {
                        ....
                    },
                    "columns": {
                        ....
                    }
                }
            )
    """

    metadata: SchemaMetadata
    columns: List[Column]

    def are_columns_the_same(
        self, new_columns: Union[List[Column], List[dict]]
    ) -> bool:
        """
        Checks that for a given Schema, does it's columns match the columns being passed
        into this function.

        Args:
            new_columns (Union[List[Column], List[dict]]): The new columns can be passed as either
                a list of Column defined classes or as a list of Python dictionaries representing
                the values. If the later is chosen and there is an incorrect value passed the function
                will raise a `rapid.exceptions.ColumnNotDifferentException`.

        Returns:
            bool: True If the new columns match the columns in the Schema otherwise False
        """
        if all(isinstance(col, dict) for col in new_columns):
            new_columns = [Column(**col) for col in new_columns]

        return self.columns == new_columns



import json
import os

import requests
from requests.auth import HTTPBasicAuth

from rapid.utils.constants import TIMEOUT_PERIOD
from rapid.exceptions import AuthenticationErrorException, CannotFindCredentialException

RAPID_CLIENT_ID = "RAPID_CLIENT_ID"
RAPID_CLIENT_SECRET = "RAPID_CLIENT_SECRET"  # pragma: allowlist secret # nosec
RAPID_URL = "RAPID_URL"


class RapidAuth:
    def __init__(
        self, client_id: str = None, client_secret: str = None, url: str = None
    ) -> None:
        """
        The rAPId auth class is a helper authentication class used to connect to your rAPId API instance. The authentication values
        can be passed into the constructor but they default to reading them from your environment variables.

        Args:
            client_id (str, optional): Your rAPId API client id token. Defaults to None.
            client_secret (str, optional): Your rAPId API client secret token. Defaults to None.
            url (str, optional): The url where your rAPId API is hosted. Defaults to None.
        """

        self.client_id = self.evaluate_inputs(client_id, RAPID_CLIENT_ID)
        self.client_secret = self.evaluate_inputs(client_secret, RAPID_CLIENT_SECRET)
        self.url = self.evaluate_inputs(url, RAPID_URL)
        self.validate_credentials()

    def evaluate_inputs(self, value: str, environment_variable: str):
        if not value:
            value = os.environ.get(environment_variable)
            if not value:
                raise CannotFindCredentialException(
                    f"No value passed for {environment_variable}, could not authenticate to rAPId"
                )
        return value

    @property
    def headers(self):
        return {"Content-Type": "application/x-www-form-urlencoded"}

    @property
    def payload(self):
        return {
            "grant_type": "client_credentials",
            "client_id": self.client_id,
        }

    def credentials_secret(self):
        return HTTPBasicAuth(self.client_id, self.client_secret)

    def request_token(self):
        response = requests.post(
            self.url + "/oauth2/token",
            auth=self.credentials_secret(),
            headers=self.headers,
            json=self.payload,
            timeout=TIMEOUT_PERIOD,
        )
        return response

    def validate_credentials(self):
        """
        Tests authentication to the rAPId API.

        Raises:
            rapid.exceptions.AuthenticationErrorException: If no authorisation can be created.

        Returns:
            None: If authentication was successful.
        """
        if self.request_token().status_code != 200:
            raise AuthenticationErrorException(
                "Auth not configured, could not connect to instance of rAPId"
            )

    def fetch_token(self):
        return json.loads(self.request_token().content.decode("utf-8"))["access_token"]



from enum import Enum
from typing import Optional, List
from pydantic.main import BaseModel


class SortDirection(Enum):
    ASC = "ASC"
    DESC = "DESC"


class SQLQueryOrderBy(BaseModel):
    column: str
    direction: SortDirection = SortDirection("ASC")


class Query(BaseModel):
    """
    A Query is a Pydantic class representing a rAPId compatible data query. It allows for programmatic definition
    of data queries. See the rAPId specific [documentation](</api/query>)
    on how to write a valid query.

    Example:
        A query can created by setting the values literally into the class like::

            query = Query(
                select_columns=["column_a", "column_b"],
                limit="5"
            )

        The alternative is you can create a schema directly from a Python dictionary::

            query = Query(
                **{
                    "select_columns": ["column_a", "column_b"],
                    "limit": "5"
                }
            )

    """

    select_columns: Optional[List[str]] = None
    filter: Optional[str] = None
    group_by_columns: Optional[List[str]] = None
    aggregation_conditions: Optional[str] = None
    order_by_columns: Optional[List[SQLQueryOrderBy]] = None
    limit: Optional[str] = None



from pandas import DataFrame
from rapid.exceptions import (
    DataFrameUploadValidationException,
    DatasetNotFoundException,
)
from rapid.items.schema import Schema, SchemaMetadata
from rapid import Rapid


def upload_and_create_dataset(
    rapid: Rapid, metadata: SchemaMetadata, df: DataFrame, upgrade_schema_on_fail=False
):
    """
    Uploads a dataframe to a dataset in the API, creating schema first if necessary.

    Args:
        rapid (Rapid): An instance of the rAPId SDK's main class.
        metadata (SchemaMetadata): The metadata for the schema to be created and the dataset to upload the DataFrame to.ß
        df (DataFrame): The pandas DataFrame to generate a schema for and upload to the dataset.
        upgrade_schema_on_fail (bool, optional): Whether to upgrade the schema if the DataFrame's schema is incorrect. Defaults to False.

    Raises:
        rapid.exceptions.DataFrameUploadValidationException: If the DataFrame's schema is incorrect and upgrade_schema_on_fail is False.
    """
    try:
        rapid.upload_dataframe(
            metadata.layer, metadata.domain, metadata.dataset, df, wait_to_complete=True
        )
    except DatasetNotFoundException:
        schema = rapid.generate_schema(
            df, metadata.layer, metadata.domain, metadata.dataset, metadata.sensitivity
        )
        schema.metadata = metadata
        rapid.create_schema(schema)
        rapid.upload_dataframe(
            metadata.layer, metadata.domain, metadata.dataset, df, wait_to_complete=True
        )
    except DataFrameUploadValidationException as validation_exception:
        if upgrade_schema_on_fail:
            update_schema_to_dataframe(rapid, metadata, df)
            rapid.upload_dataframe(
                metadata.layer,
                metadata.domain,
                metadata.dataset,
                df,
                wait_to_complete=True,
            )
        else:
            raise validation_exception


def update_schema_to_dataframe(
    rapid: Rapid,
    metadata: SchemaMetadata,
    df: DataFrame,
):
    """
    Updates a schema for a specified dataset in the API to match the given Dataframe.

    Args:
        rapid (Rapid): An instance of the rAPId SDK's main class.
        metadata (SchemaMetadata): The metadata for the schema to be updated and the dataset the DataFrame belongs to.
        df (Dataframe): The dataframe that the schema should be updated to match.
    """
    schema_response = rapid.generate_schema(
        df, metadata.layer, metadata.domain, metadata.dataset, metadata.sensitivity
    )
    schema = Schema(metadata=metadata, columns=schema_response.columns)
    rapid.update_schema(schema)



from setuptools import setup, find_packages

setup(
    name="rapid-sdk",
    version="0.1.5",
    description="A python sdk for the rAPId API",
    url="https://github.com/no10ds/rapid-sdk",
    author="Lewis Card",
    author_email="lcard@no10.gov.uk",
    license="MIT",
    packages=find_packages(include=["rapid", "rapid.*"], exclude=["tests"]),
    install_requires=["pandas", "requests", "deepdiff"],
    include_package_data=True,
)



import json
import time
import requests

from datetime import datetime
from typing import Dict, Optional

import pandas as pd

from pandas import DataFrame

from rapid.auth import RapidAuth
from rapid.items.schema import Schema
from rapid.items.query import Query
from rapid.utils.constants import TIMEOUT_PERIOD
from rapid.exceptions import (
    DataFrameUploadFailedException,
    DataFrameUploadValidationException,
    JobFailedException,
    SchemaGenerationFailedException,
    SchemaCreateFailedException,
    SchemaUpdateFailedException,
    SchemaAlreadyExistsException,
    UnableToFetchJobStatusException,
    DatasetInfoFailedException,
    DatasetNotFoundException,
    InvalidPermissionsException,
    SubjectAlreadyExistsException,
    SubjectNotFoundException,
)


class Rapid:
    def __init__(self, auth: RapidAuth = None) -> None:
        """
        The rAPId class is the main SDK class for the rAPId API. It acts as a wrapper for the various
        API endpoints, providing a simple and intuitive programmatic interface.

        Args:
            auth (int, optional): An instance of the rAPId auth class, which is used for authentication and authorization with the API. Defaults to None.
        """
        self.auth = auth if auth else RapidAuth()

    def generate_headers(self) -> Dict:
        return {"Authorization": "Bearer " + self.auth.fetch_token()}

    def list_datasets(self):
        """
        List all current datasets within rAPId instance.

        Returns:
            A JSON response of the API's response.
        """
        response = requests.post(
            f"{self.auth.url}/datasets",
            headers=self.generate_headers(),
            timeout=TIMEOUT_PERIOD,
        )
        return json.loads(response.content.decode("utf-8"))

    def fetch_job_progress(self, _id: str):
        """
        Makes a GET request to the API to fetch the progress of a specific job.

        Args:
            _id (str): The ID of the job to fetch the progress for.

        Returns:
            A JSON response of the API's response.
        """
        url = f"{self.auth.url}/jobs/{_id}"
        response = requests.get(
            url, headers=self.generate_headers(), timeout=TIMEOUT_PERIOD
        )
        data = json.loads(response.content.decode("utf-8"))
        if response.status_code == 200:
            return data
        raise UnableToFetchJobStatusException("Could not check job status", data)

    def wait_for_job_outcome(self, _id: str, interval: int = 1):
        """
        Makes periodic requests to the API to wait for the outcome of a specific job.

        Args:
            _id (str): The ID of the job to wait for the outcome of.
            interval (int, optional): The number of seconds to sleep between requests to the API. Defaults to 1.

        Returns:
            None if the job is successful.

        Raises:
            rapid.exceptions.JobFailedException: If the job outcome failed.
        """
        while True:
            progress = self.fetch_job_progress(_id)
            status = progress["status"]
            if status == "SUCCESS":
                return None
            if status == "FAILED":
                raise JobFailedException("Upload failed", progress)
            time.sleep(interval)

    def download_dataframe(
        self,
        layer: str,
        domain: str,
        dataset: str,
        version: Optional[int] = None,
        query: Query = Query(),
    ) -> DataFrame:
        """
        Downloads data to a pandas DataFrame based on the domain, dataset and version passed.

        Args:
            layer (str): The layer of the dataset to download the DataFrame from.
            domain (str): The domain of the dataset to download the DataFrame from.
            dataset (str): The dataset from the domain to download the DataFrame from.
            version (int, optional): Version of the dataset to download.
            query (rapid.items.query.Query, optional): An optional query type to provide when downloading data. Defaults to empty.

        Raises:
            DatasetNotFoundException: rapid.exceptions.DatasetNotFoundException: If the
                specificed domain, dataset and version to download does not exist in the rAPId instance
                we throw the dataset not found exception.

        Returns:
            DataFrame: A pandas DataFrame of the data
        """
        url = f"{self.auth.url}/datasets/{layer}/{domain}/{dataset}/query"
        if version is not None:
            url = f"{url}?version={version}"
        response = requests.post(
            url,
            headers=self.generate_headers(),
            data=json.dumps(query.dict(exclude_none=True)),
            timeout=TIMEOUT_PERIOD,
        )
        data = json.loads(response.content.decode("utf-8"))
        if response.status_code == 200:
            return pd.read_json(json.dumps(data), orient="index")

        raise DatasetNotFoundException(
            f"Could not find dataset, {layer}/{domain}/{dataset} to download", data
        )

    def upload_dataframe(
        self,
        layer: str,
        domain: str,
        dataset: str,
        df: DataFrame,
        wait_to_complete: bool = True,
    ):
        """
        Uploads a pandas DataFrame to a specified dataset in the API.

        Args:
            layer (str): The layer of the dataset to upload the DataFrame to.
            domain (str): The domain of the dataset to upload the DataFrame to.
            dataset (str): The name of the dataset to upload the DataFrame to.
            df (DataFrame): The pandas DataFrame to upload.
            wait_to_complete (bool, optional): Whether to wait for the upload job to complete before returning. Defaults to True.

        Raises:
            rapid.exceptions.DataFrameUploadValidationException: If the DataFrame's schema is incorrect.
            rapid.exceptions.DataFrameUploadFailedException: If an unexpected error occurs while uploading the DataFrame.
            rapid.exceptions.DatasetNotFoundException: If the specified dataset does not exist.

        Returns:
            If wait_to_complete is True, returns "Success" if the upload is successful.
            If wait_to_complete is False, returns the ID of the upload job if the upload is accepted.
        """
        url = f"{self.auth.url}/datasets/{layer}/{domain}/{dataset}"
        response = requests.post(
            url,
            headers=self.generate_headers(),
            files=self.convert_dataframe_for_file_upload(df),
            timeout=TIMEOUT_PERIOD,
        )
        data = json.loads(response.content.decode("utf-8"))

        if response.status_code == 202:
            if wait_to_complete:
                self.wait_for_job_outcome(data["details"]["job_id"])
                return "Success"
            return data["details"]["job_id"]
        if response.status_code == 422:
            raise DataFrameUploadValidationException(
                "Could not upload dataframe due to an incorrect schema definition"
            )
        elif response.status_code == 404:
            raise DatasetNotFoundException(
                "Could not find dataset: {layer}/{domain}/{dataset}", data
            )
        else:
            raise DataFrameUploadFailedException(
                "Encountered an unexpected error, could not upload dataframe",
                data["details"],
            )

    def fetch_dataset_info(self, layer: str, domain: str, dataset: str):
        """
        Fetches information about the specified dataset in the API.

        Args:
            layer (str): The layer of the dataset to fetch metadata for.
            domain (str): The domain of the dataset to fetch metadata for.
            dataset (str): The name of the dataset to fetch metadata for.

        Raises:
            rapid.exceptions.DatasetInfoFailedException: If an error occurs while fetching the dataset information.
            rapid.exceptions.DatasetNotFoundException: If the specified dataset does not exist.

        Returns:
            A dictionary containing the metadata information for the dataset.
        """
        url = f"{self.auth.url}/datasets/{layer}/{domain}/{dataset}/info"
        response = requests.get(
            url,
            headers=self.generate_headers(),
            timeout=TIMEOUT_PERIOD,
        )
        data = json.loads(response.content.decode("utf-8"))
        if response.status_code == 200:
            return data

        if response.status_code == 404:
            raise DatasetNotFoundException(
                f"Could not find dataset, {layer}/{domain}/{dataset} to get info", data
            )

        raise DatasetInfoFailedException(
            "Failed to gather the dataset info", data["details"]
        )

    def convert_dataframe_for_file_upload(self, df: DataFrame):
        """
        Converts a pandas DataFrame to a format that can be used for file uploads to the API.

        Args:
            df (DataFrame): The pandas DataFrame to convert.

        Returns:
            A dictionary containing the converted DataFrame in a format suitable for file uploads to the API.
        """
        return {
            "file": (
                f"rapid-sdk-{int(datetime.now().timestamp())}.csv",
                df.to_csv(index=False),
            )
        }

    def generate_schema(
        self, df: DataFrame, layer: str, domain: str, dataset: str, sensitivity: str
    ) -> Schema:
        """
        Generates a schema for a pandas DataFrame and a specified dataset in the API.

        Args:
            df (DataFrame): The pandas DataFrame to generate a schema for.
            layer (str): The layer of the dataset to generate a schema for.
            domain (str): The domain of the dataset to generate a schema for.
            dataset (str): The name of the dataset to generate a schema for.
            sensitivity (str): The sensitivity level of the schema to generate.

        Raises:
            rapid.exceptions.SchemaGenerationFailedException: If an error occurs while generating the schema.

        Returns:
            rapid.items.schema.Schema: A Schema class type from the generated schema for the DataFrame and dataset.
        """
        url = (
            f"{self.auth.url}/schema/{layer}/{sensitivity}/{domain}/{dataset}/generate"
        )

        response = requests.post(
            url,
            headers=self.generate_headers(),
            files=self.convert_dataframe_for_file_upload(df),
            timeout=TIMEOUT_PERIOD,
        )
        data = json.loads(response.content.decode("utf-8"))
        if response.status_code == 200:
            return Schema(**data)
        raise SchemaGenerationFailedException("Could not generate schema", data)

    def create_schema(self, schema: Schema):
        """
        Creates a new schema on the API.

        Args:
            schema rapid.items.schema.Schema: The schema model for which you want to create for.

        Raises:
            rapid.exceptions.SchemaAlreadyExistsException: If you try to create a schema that already exists in rAPId.
            rapid.exceptions.SchemaCreateFailedException: If an error occurs while trying to update the schema.
        """
        schema_dict = schema.dict()
        url = f"{self.auth.url}/schema"
        response = requests.post(
            url,
            headers=self.generate_headers(),
            data=json.dumps(schema_dict),
            timeout=TIMEOUT_PERIOD,
        )
        if response.status_code == 201:
            pass
        elif response.status_code == 409:
            raise SchemaAlreadyExistsException("The schema already exists")
        else:
            data = json.loads(response.content.decode("utf-8"))
            raise SchemaCreateFailedException("Could not create schema", data)

    def update_schema(self, schema: Schema):
        """
        Uploads a new updated schema to the API.

        Args:
            schema (rapid.items.schema.Schema): The new schema model that will be used for the update.

        Raises:
            rapid.exceptions.SchemaUpdateFailedException: If an error occurs while trying to update the schema.
        """
        schema_dict = schema.dict()
        url = f"{self.auth.url}/schema"
        response = requests.put(
            url,
            headers=self.generate_headers(),
            data=json.dumps(schema_dict),
            timeout=TIMEOUT_PERIOD,
        )
        data = json.loads(response.content.decode("utf-8"))
        if response.status_code == 200:
            return data
        raise SchemaUpdateFailedException("Could not update schema", data)

    def create_client(self, client_name: str, client_permissions: list[str]):
        """
        Creates a new client on the API with the specified permissions.

        Args:
            client_name (str): The name of the client to create.
            client_permissions (list[str]): The permissions of the client to create.

        Raises:
            rapid.exceptions.InvalidPermissionsException: If an error occurs while trying to create the client.
        """
        url = f"{self.auth.url}/client"
        response = requests.post(
            url,
            headers=self.generate_headers(),
            data=json.dumps(
                {"client_name": client_name, "permissions": client_permissions}
            ),
            timeout=TIMEOUT_PERIOD,
        )
        data = json.loads(response.content.decode("utf-8"))
        if response.status_code == 201:
            return data
        elif response.status_code == 400:
            raise SubjectAlreadyExistsException(
                f"The client {client_name} already exists"
            )
        raise InvalidPermissionsException(
            "One or more of the provided permissions is invalid or duplicated"
        )

    def delete_client(self, client_id: str) -> None:
        """
        Deletes a client from the API based on their id

        Args:
            client_id (str): The id of the client to delete.

        Raises:
            rapid.exceptions.SubjectNotFoundException: If the client does not exist.
        """
        url = f"{self.auth.url}/client/{client_id}"
        response = requests.delete(
            url,
            headers=self.generate_headers(),
            timeout=TIMEOUT_PERIOD,
        )
        if response.status_code == 200:
            return None

        raise SubjectNotFoundException(
            f"Failed to delete client with id: {client_id}, ensure it exists."
        )

    def update_subject_permissions(self, subject_id: str, permissions: list[str]):
        """
        Updates the permissions of a subject on the API.

        Args:
            subject_id (str): The id of the subject to update.
            permissions (list[str]): The permissions to update the subject with.

        Raises:
            rapid.exceptions.InvalidPermissionsException: If an error occurs while trying to create the client.
        """
        url = f"{self.auth.url}/subject/permissions"
        response = requests.put(
            url,
            headers=self.generate_headers(),
            data=json.dumps({"subject_id": subject_id, "permissions": permissions}),
            timeout=TIMEOUT_PERIOD,
        )
        data = json.loads(response.content.decode("utf-8"))
        if response.status_code == 200:
            return data

        raise InvalidPermissionsException(
            "One or more of the provided permissions is invalid or duplicated"
        )
